import numpy as np
from collections import Counter

class RandomForest:
    def __init__(self, n_trees=10, k=None, max_depth=10, min_samples_split=2):
        self.n_trees = n_trees
        self.k = k
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.trees = []

    def fit(self, X, y):
        self.trees = []
        for _ in range(self.n_trees):
            tree = self._build_tree(X, y)
            X_sample, y_sample = self._bootstrap_samples(X, y)
            tree.fit(X_sample, y_sample)
            self.trees.append(tree)

    def _build_tree(self, X, y):
        return self._DecisionTree(max_depth=self.max_depth, 
                                  min_samples_split=self.min_samples_split, 
                                  n_features=self.k)

    def _bootstrap_samples(self, X, y):
        n_samples = X.shape[0]
        idxs = np.random.choice(n_samples, n_samples, replace=True)
        return X[idxs], y[idxs]

    def predict(self, X):
        predictions = [self._predict(x) for x in X]
        return predictions

    def _predict(self, x):
        tree_preds = np.array([tree.predict([x])[0] for tree in self.trees])
        most_common = Counter(tree_preds).most_common()
        return most_common[0][0]

    class _DecisionTree:
        def __init__(self, max_depth=10, min_samples_split=2, n_features=None):
            self.max_depth = max_depth
            self.min_samples_split = min_samples_split
            self.n_features = n_features
            self.root = None

        def fit(self, X, y):
            self.root = self._grow_tree(X, y)

        def _grow_tree(self, X, y, depth=0):
            n_samples, n_feats = X.shape
            n_labels = len(np.unique(y))

            if depth >= self.max_depth or n_labels == 1 or n_samples < self.min_samples_split:
                leaf_value = self._most_common_label(y)
                return self._Node(value=leaf_value)

            feat_idxs = np.random.choice(n_feats, self.n_features, replace=False)
            best_feature, best_thresh = self._best_split(X, y, feat_idxs)

            left_idxs, right_idxs = self._split(X[:, best_feature], best_thresh)
            left = self._grow_tree(X[left_idxs, :], y[left_idxs], depth + 1)
            right = self._grow_tree(X[right_idxs, :], y[right_idxs], depth + 1)
            return self._Node(best_feature, best_thresh, left, right)

        def _best_split(self, X, y, feat_idxs):
            best_gain = -1
            split_idx, split_threshold = None, None

            for feat_idx in feat_idxs:
                X_column = X[:, feat_idx]
                thresholds = np.unique(X_column)

                for thr in thresholds:
                    gain = self._information_gain(y, X_column, thr)

                    if gain > best_gain:
                        best_gain = gain
                        split_idx = feat_idx
                        split_threshold = thr

            return split_idx, split_threshold

        def _information_gain(self, y, X_column, threshold):
            parent_entropy = self._entropy(y)

            left_idxs, right_idxs = self._split(X_column, threshold)

            if len(left_idxs) == 0 or len(right_idxs) == 0:
                return 0

            n = len(y)
            n_l, n_r = len(left_idxs), len(right_idxs)
            e_l, e_r = self._entropy(y[left_idxs]), self._entropy(y[right_idxs])
            child_entropy = (n_l / n) * e_l + (n_r / n) * e_r

            information_gain = parent_entropy - child_entropy
            return information_gain

        def _split(self, X_column, split_thresh):
            left_idxs = np.argwhere(X_column <= split_thresh).flatten()
            right_idxs = np.argwhere(X_column > split_thresh).flatten()
            return left_idxs, right_idxs

        def _entropy(self, y):
            hist = np.bincount(y)
            ps = hist / len(y)
            return -np.sum([p * np.log(p) for p in ps if p > 0])

        def _most_common_label(self, y):
            counter = Counter(y)
            return counter.most_common(1)[0][0]

        def predict(self, X):
            return np.array([self._traverse_tree(x, self.root) for x in X])

        def _traverse_tree(self, x, node):
            if node.is_leaf_node():
                return node.value

            if x[node.feature] <= node.threshold:
                return self._traverse_tree(x, node.left)
            return self._traverse_tree(x, node.right)

        class _Node:
            def __init__(self, feature=None, threshold=None, left=None, right=None, *, value=None):
                self.feature = feature
                self.threshold = threshold
                self.left = left
                self.right = right
                self.value = value

            def is_leaf_node(self):
                return self.value is not None
